/*****************************************************************************

 \G P C M/                                                          \G P C M/
 Ŀ             GPCM ESCC DMA card test program              Ŀ
 ͵Ĵ            Rev. 2.1 - (C) 1997,1998,2000 GPCM            ͵Ĵ
 ( )Ĵ                                                          ( )Ĵ
 ͵       This program is FREE SOFTWARE and comes with       ͵
 `.' ABSOLUTELY NO WARRANTY: see the file COPYING for details `.'

 *****************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <dos.h>
#include "85230.h"
#include "8254.h"
#include "gpcmtest.h"

#undef DEBUG
void interrupt far isr3(void);
void interrupt far isr4(void);
void interrupt far isr5(void);
void interrupt far isr7(void);
void interrupt far isr8(void);
void interrupt far isr9(void);
void interrupt far isr10(void);
void interrupt far isr11(void);

int have_8254, have_scc, have_escc, int_mask=0;
unsigned int irq=0;
void interrupt far (*isrs[16])() = {
	NULL,NULL,NULL, isr3, isr4, isr5, NULL,isr7,
	NULL,isr9,isr10,isr11,NULL,NULL,NULL,NULL
};

int test_address(int base,char *reason)
{
	int n, NotFound=1;

	printf("Checking for I/O ports at 0x%x and 0x%x................", base, base + TMR0);

	outp(base + TMRCMD, SC0 | LSB_MSB | MODE3);

	for (n=0; n<10; n++) {
		NotFound &= (inp(base) == 0xff);
		NotFound &= (inp(base+TMR0) == 0xff);
	}
	
	if (NotFound) {
		printf("FAILED\n");
		strcat(reason,"\n\tCard not found, check address and access delay");
		return 0;
	} else {
		printf("PASSED\n");
		return 1;
	}
}


int test_825X(int base,char *reason)
{
	int n;
	unsigned int count1, count2;
	unsigned char status1,status2;

	printf("Checking for 8253/8254 timer at 0x%x....................",base+TMR0);

	/* Set up counter chip, mode 3,  */
	outp(base + TMRCMD, SC0 | LSB_MSB | MODE3);
	delay(5);
	outp(base + TMR0, 0xff);
	delay(2);
	outp(base + TMR0, 0xff);
	delay(2);

	/* 
	 * Try to read status using the "Read Back" feature.
	 *
	 * Mode read must be equal to the mode selected above.
	 * That works only with a 8254 chip.
	 *
	 * Status is read two times to ensure 8253s are always
	 * correctly detected (on 8253s, counter 0 is actually
	 * read, not status, so reading it only once status may
	 * be equal to mode by chance, resulting in a false 8254
	 * detection.
	 */
	outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC0);
	delay(2);

	status1 = inp(base + TMR0);

	outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC0);
	delay(2);

	status2 = inp(base + TMR0);

	delay(2);
#ifdef DEBUG
	printf("\nDEBUG: status1=0x%2x, status2=0x%2x\n",status1,status2);
#endif
	if (((status1 & MODE) == (LSB_MSB | MODE3)) && ((status2 & MODE) == (LSB_MSB | MODE3))) {
		/* Mode reads back OK, this is a 8254 */
		have_8254 = 1;
	} else {
		/* This is NOT a 8254 */
		have_8254 = 0;
	}

	/*
	 * Try to latch and read counter 0 two times using the
	 * "Latch count" feature.
	 * That must work on all good 8253's & 8254's.
	 */
	n=0;
	do {
		/* 
		 * Wait for a count > 49152 to ensure counter 0
		 * will not roll between first and second read.
		 *
		 * Timeout after 10000 unsuccesful reads.
		 */
		outp(base + TMRCMD, CLATCH | SC0);
		count1 = inp(base + TMR0);
		count1 += (inp(base + TMR0) << 8);
	} while ((count1 < 49152) && (n++ < 10000));

	/*
	 * Latch and read counter 0 again
	 */
	outp(base + TMRCMD, CLATCH | SC0);
	count2 = inp(base + TMR0);
	count2 += (inp(base + TMR0) << 8);

#ifdef DEBUG
	printf("\nDEBUG: count1=%u, count2=%u\n",count1,count2);
#endif
	/*
	 * Since the 8253/8254 counters are DOWN COUNTERS,
	 * count2 must be LESS than count1 for a good chip
	 */

	if (count1 > count2) {
		printf("PASSED (%s found)\n",have_8254 ? "8254" : "8253");
		return 1;
	} else {
		printf("FAILED\n");
		strcat(reason,"\n\tTimer chip not found");
		return 0;
	}
}


int test_escc(int base,char *reason)
{
	int ctla, ctlb;

	printf("Checking for 85230 ESCC chip at 0x%x....................",base);
	
	ctla = base + CHANA + CTL; /* ESCC channel A control register */
	ctlb = base + CHANB + CTL; /* ESCC channel B control register */

	/*
	 * Reset 85230
	 */
	 outp(ctla, R9);
	 outp(ctla, FHWRES); 
	 delay(50);
	
	/*
	 * Disable interrupts
	 */
	 outp(ctla, R9);
	 outp(ctla, 0);
	 
	/*
	 * check for ESCC
	 */
	outp(ctlb, R15);
	outp(ctlb,0);
	
	outp(ctlb,R15);
	have_scc =  ((inp(ctlb) & ESCC) == 0);

	if (have_scc) {
		outp(ctlb, R15);
		outp(ctlb, ESCC);
		outp(ctlb, R15);
		have_escc = (inp(ctlb) & ESCC);
		outp(ctlb, R15);
		outp(ctlb, 0);
		printf("PASSED (%s found)\n",have_escc ? "85230" : "8530");
	} else {
		printf("FAILED\n");
		strcat(reason,"\n\t(E)SCC chip not found");
		return 0;
	}
	return 1;
}

int test_825X_and_ESCC(int base, char *reason)
{
	int ctla, ctlb, out1, out2, ctsa, ctsb,
	    test_ok, i8254_ok, escc_ok, clock_ok;

	printf("Performing %s/%s final test.........................",
		have_8254 ? "8254" : "8253",
		have_escc ? "85230" : "8530 "); 

	/*
	 * Program TIMER0 for 4000 Hz square wave output
	 * to TIMER1 & TIMER2
	 */
	outp(base+TMRCMD, SC0 | LSB_MSB | MODE3);
	delay(2);
	outp(base+TMR0, DIV0 & 0xff);
	delay(2);
	outp(base+TMR0, (DIV0 >> 8) & 0xff);
	
	/*
	 * Program TIMER1 & TIMER2 in mode 0
	 */
	outp(base+TMRCMD, SC1 | LSB_MSB | MODE0);
	delay(2);
	outp(base+TMRCMD, SC2 | LSB_MSB | MODE0);
	delay(2);

	/*
	 * Load 100 ms count (400 0.25 ms units) in TIMER 1 & TIMER 2
	 */
	outp(base+TMR1,0x90);
	delay(2);
	outp(base+TMR1,0x1);
	delay(2);
	outp(base+TMR2,0x90);
	delay(2);
	outp(base+TMR2,0x1);
	
	if (have_8254) {
		/* 
		 * If we have a 8254, read status using the "Read Back"
		 * feature. Status bit 7 is the OUT value.
		 * After counters loading OUT value must be 0.
		 */
		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC1);
		delay(2);
		out1 = (inp(base + TMR1) & OUT) >> 7;
		delay(2);
	
		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC2);
		delay(2);
		out2 = inp(base + TMR2) & OUT;
		
		test_ok = ((out1 == 0) && (out2 == 0));
#ifdef DEBUG
		printf("\nDEBUG: out1=%d, out2=%d\n", out1, out2);
#endif
		/*
		 * Now wait for slightly less than 100 ms
		 * and verify if the OUT pin is still 0 
		 */
		delay(80);
		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC1);
		delay(2);
		out1 = inp(base + TMR1) & OUT;
		delay(2);

		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC2);
		delay(2);
		out2 = inp(base + TMR2) & OUT;  

		test_ok &= ((out1 == 0) && (out2 == 0));
#ifdef DEBUG
		printf("\nDEBUG: after 90 ms: out1=%d, out2=%d\n", out1, out2);

#endif

		/*
		 * Now wait for another 20 ms and verify if
		 * the OUT pin is now 1
		 */
		delay(20);
		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC1);
		delay(2);
		out1 = inp(base + TMR1) & OUT;
		delay(2);

		outp(base + TMRCMD, RDBACK | RDBSTA | RDBSC2);
		delay(2);
		out2 = inp(base + TMR2) & OUT;  
#ifdef DEBUG
		printf("\nDEBUG: after 110 ms: out1=%d, out2=%d\n", out1, out2);

#endif
		test_ok &= ((out1 != 0) && (out2 != 0));
		i8254_ok = test_ok;
	} else {
		/* 
		 * No way to read OUT pin status, assume timer chip OK
		 */
		i8254_ok = test_ok = 1;
	}


	/*
	 * Init 85230 port A & B control address
	 */
	ctla = base + CHANA + CTL; /* ESCC channel A control register */
	ctlb = base + CHANB + CTL; /* ESCC channel B control register */
	
	/*
	 * Re-init 85230 channel A & B
	 */
	 outp(ctla,R15);
	 outp(ctla,0);
	 outp(ctlb,R15);
	 outp(ctlb,0);

	/*
	 * Load 100 ms count (400 0.25 ms units) in TIMER 1 & TIMER 2
	 */
	outp(base+TMR1,0x90);
	delay(2);
	outp(base+TMR1,0x1);
	delay(2);
	outp(base+TMR2,0x90);
	delay(2);
	outp(base+TMR2,0x1);
	delay(2);

	/*
	 * Now read OUT1 & OUT2 bits from 85230 CTSB & CTSA
	 */
	outp(ctlb, R0);
	ctsb = inp(ctlb) & CTS;
	outp(ctla, R0);
	ctsa = inp(ctla) & CTS;
	escc_ok = ((ctsb != 0) && (ctsa != 0));
	test_ok &= escc_ok;
#ifdef DEBUG
	printf("\nDEBUG: ctsb=%d, ctsa=%d\n", ctsb, ctsa);
#endif
	/*
	 * Now wait for slightly less than 100 ms
	 * and verify if the OUT pin is still 0 
	 */
	delay(90);
	
	outp(ctlb, R0);
	ctsb = inp(ctlb) & CTS;
	outp(ctla, R0);
	ctsa = inp(ctla) & CTS;
	clock_ok = ((ctsb != 0) && (ctsa != 0));
#ifdef DEBUG
	printf("\nDEBUG: after 90 ms: ctsb=%d, ctsa=%d\n", ctsb, ctsa);
#endif
	/*
	 * Now wait for another 20 ms
	 * and verify if the OUT pin is now 1 
	 */
	delay(20);
	
	outp(ctlb, R0);
	ctsb = inp(ctlb) & CTS;
	outp(ctla, R0);
	ctsa = inp(ctla) & CTS;
	escc_ok &= ((ctsb == 0) && (ctsa == 0));
	test_ok &= escc_ok;
#ifdef DEBUG
	printf("\nDEBUG: after 110 ms: ctsb=%d, ctsa=%d\n", ctsb, ctsa);
#endif

	if (i8254_ok == 0) {
		strcat(reason,"\n\t8254 timer failure");
	}
	
	if (clock_ok == 0) {
		strcat(reason, ": timer clock is too fast, check 4.91 MHz oscillator");
	}
	
	if (escc_ok == 0) {
		strcat(reason,"\n\t(E)SCC CTS read failure");
	}
	
	if (test_ok) {
		printf("PASSED\n");
	} else {
		printf("FAILED\n");
	}

	return test_ok;
}

/*
 * Dummy interrupt service routines (ISRs) for IRQ autoprobing.
 * Set irq to IRQ value, issue an EOI to PICs and return.
 * IRQs 0,1,2,8,14 and 15 are already in use, IRQ 12 and 13 may
 * be available, but are not directly supported by the card, so
 * are not probed for.
 * Note that the IRQ2 line on AT's is really IRQ 9, since IRQ 2 is
 * used to cascade the second PIC.
 */
void interrupt far isr3(void)
{
	irq = 3;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr4(void)
{
	irq = 4;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr5(void)
{
	irq = 5;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr7(void)
{
	irq = 7;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr8(void)
{
	irq = 8;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr9(void)
{
	irq = 9;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr10(void)
{
	irq = 10;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

void interrupt far isr11(void)
{
	irq = 11;
	outp(0xa0, 0x20);
	outp(0x20, 0x20);
}

/*
 * Try to autobrobe IRQ used.
 * Get all free IRQs, program the card to issue an interrupt
 * and see on what IRQ line it is coming up.
 */
int probe_irq(int base, char *reason)
{
	void interrupt far (*oldvect[16])();
	int ctla = base + CHANA + CTL;
	int i;
	int ctsa1, ctsa2;

	printf("Trying to autoprobe IRQ vector...........................");

	/* Get and store old vectors, set new vectors and enable interrupts */
	for (i=0; i<16; i++) {
		if (i < 8) {
			oldvect[i] = getvect(i+8);
			if (isrs[i] != NULL)
				setvect(i+8, isrs[i]);
		} else {
			oldvect[i] = getvect(0x70+i-8);
			if (isrs[i] != NULL)
				setvect(0x70+i-8, isrs[i]);
		}
	}
	/*
	 * Reset 85230
	 */
	outp(base+DMAEN, 0);
	outp(ctla, R9);
	outp(ctla, FHWRES);
	delay(50);
	/* 
	 * Enable GPCM card interrupts
	 */
	outp(ctla, R9);
	outp(ctla, MIE | NV);
	delay(1);
	outp(ctla, R1);
	outp(ctla, EXT_INT_ENAB);
	delay(1);
	outp(ctla, R15);
	outp(ctla, CTSIE);
	delay(1);
	/*
	 * Program TIMER0 for 4000 Hz square wave output
	 * to TIMER1 & TIMER2
	 */
	outp(base+TMRCMD, SC0 | LSB_MSB | MODE3);
	delay(1);
	outp(base+TMR0, DIV0 & 0xff);
	delay(1);
	outp(base+TMR0, (DIV0 >> 8) & 0xff);
	delay(1);
	/*
	 * Enable all interrupts in PICs
	 */
	outp(0x21,0);
	outp(0xa1,0);
	outp(base+DMAEN, 1);

	/*
	 * Program TIMER1 in mode 0
	 */
	outp(base+TMRCMD, SC1 | LSB_MSB | MODE0);
	delay(1);
	/*
	 * Load 50 ms count (200 0.25 ms units) in TIMER 1
	 */
	outp(base+TMR1,200);
	delay(1);
	outp(base+TMR1,0);
	delay(1);
	/*
	 * Reset external interrupt
	 */
	outp(ctla, RES_EXT_INT);
	delay(1);
	outp(ctla, RES_EXT_INT);
	delay(1);

#ifdef DEBUG
	outp(ctla, R3);
	i=inp(ctla);
	printf("After RES_EXT_INT, int status = %2.2x\n", i);
#endif

	/* 
	 * Wait 100 ms for the interrupt to occur
	 */
	delay(100);

#ifdef DEBUG
	outp(ctla, R3);
	i=inp(ctla);
	printf("After 100 ms, int status = %2.2x\n", i);
#endif
	/*
	 * Restore original vectors
	 */
	for (i=0; i<16; i++) {
		if (i < 8) {
			setvect(i+8, oldvect[i]);
		} else {
			setvect(0x70+i-8, oldvect[i]);
		}
	}
	/*
	 * Announce results
	 */
	if (irq == 0) {
		printf("FAILED\n");
		strcat(reason,"\n\tunable to detect IRQ vector.");
	} else {
		printf("PASSED (IRQ %d found)\n", irq);
	}
	return irq;
}

/*
 * Try to autoprobe the DMA channel used.
 * Read DMA status register, then trigger a DMA request, re-read
 * the DMA status register looking for a DMA request bit set.
 */
int probe_dma(int base, char *reason)
{
	int ctla = base + CHANA + CTL;
	short dma_status, new_dma_status;
	int dma = 0, i;

	printf("Trying to autoprobe DMA channel..........................");
#if 0
	/* Disable DMA */
	outp(base+DMAEN,0);
#endif			
	/* Read the DMA channel status registers. */
	dma_status = ((inp(DMA_STAT) >> 4) & 0x0f);
#ifdef DEBUG
	printf("DMA status = %d\n",dma_status);
#endif
	outp(base+DMAEN, 3);

	/* Trigger a DMA request, perhaps pause a bit. */
	outp(ctla, R1);
	delay(1);
	outp(ctla, WT_RDY_ENAB | WT_FN_RDYFN);

	delay(50);

	outp(base+DMAEN, 0);

	/* Re-read the DMA status registers. */
	new_dma_status = ((inp(DMA_STAT) >> 4) & 0x0f);
#ifdef DEBUG
	printf("new DMA status = %d\n",new_dma_status);
#endif
	/*
	 * Eliminate the old and floating requests
	 * and search for a new DMA request
	 */
	new_dma_status ^= dma_status;
	for (i = 3; i > 0; i--) {
		if ((new_dma_status & (1<<i))) {
			dma = i;
			break;
		}
	}
	if (i <= 0) {
		printf("FAILED\n");
		strcat(reason,"\n\tunable to detect DMA channel.");
	} else {
		printf("PASSED (DMA %d found)\n", dma);
	}
	
	outp(base + DMAEN, 0);

	return dma;
}

main(int argc, char *argv[])

{
	int base, card_ok, irq_no, dma_no;
	char reason[255];

	clrscr();
	
	printf("Ŀ\n");
	printf("                 GPCM ESCC DMA CARD TEST PROGRAM - Rev. 2.1                 \n");
	printf("\n\n");

	if (argc < 2) {
		printf("Usage: gpcmtest <base address>\n");
		printf("<base address> must be in the form 0xNNN, e. g.: gpcmtest 0x380\n\n");
		printf("No base address given, using default (0x380)\n\n");
		base = 0x380;
	} else {
		sscanf(argv[1], "0x%x", &base);
	}
	*reason = '\0';
	printf("Testing for GPCM ESCC card at 0x%x:\n\n", base);
	
	card_ok = test_address(base,reason);

	if (card_ok) {
		/*
		 * A card is responding at the specified I/O address
		 * tests can continue
		 */
		card_ok &= test_825X(base, reason);
		card_ok &= test_escc(base, reason);
	}

	if (card_ok) {
		/*
		 * Both 8254 and ESCC available: do final test.
		 */
		 card_ok &= test_825X_and_ESCC(base, reason);
		 /*
		  * Try to autodetect IRQ and DMA used.
		  */
		 printf("-----------------------------------------------------------------------------\n");
		 printf("            Now we try to autodetect IRQ and DMA used by the card            \n");
		 printf("THIS IS ONLY AN ATTEMPT: USUALLY WORKS, BUT DON'T TRUST 100%% NEGATIVE RESULTS\n"); 
		 printf("-----------------------------------------------------------------------------\n");
		 irq_no = probe_irq(base, reason);
		 dma_no = probe_dma(base, reason);
	}

	if (card_ok) {
		printf("\n\nTest passed: GPCM ESCC card at 0x%x is OK.\n",base);
		if (irq_no) {
			printf("             GPCM ESCC card: autodetected IRQ %d.\n",irq_no);
		}
		if (dma_no) {
			printf("             GPCM ESCC card: autodetected DMA %d.\n",dma_no);
		}
		if (have_escc == 0) {
			printf("\nWARNING: standard SCC (8530) found, NOS driver code REQUIRES a ESCC (85230) !\n");
		}
	} else {
		printf("\n\nTest failed due to the following reason(s): %s\n\n", reason);
	}
	return 0;
}

